@@ -25,36 +25,33 @@
 #include <config.h>
 #endif
 
-#include <ctype.h>
 #include <unistd.h>
-#include <stdio.h>
-#include <string.h>
-#include <stdlib.h>
 #include <errno.h>
 #include <netdb.h>
 #include <time.h>
+
 #include <sys/socket.h>
 #include <sys/mount.h>
+#include <netinet/in.h>
+#include <arpa/inet.h>
 
 #include "xcommon.h"
 #include "mount.h"
 #include "nls.h"
-#include "nfs_mount.h"
 #include "mount_constants.h"
 #include "stropts.h"
 #include "error.h"
 #include "network.h"
 #include "parse_opt.h"
+#include "version.h"
+#include "parse_dev.h"
 
-#ifdef HAVE_RPCSVC_NFS_PROT_H
-#include <rpcsvc/nfs_prot.h>
-#else
-#include <linux/nfs.h>
-#define nfsstat nfs_stat
+#ifndef NFS_PROGRAM
+#define NFS_PROGRAM	(100003)
 #endif
 
 #ifndef NFS_PORT
-#define NFS_PORT 2049
+#define NFS_PORT	(2049)
 #endif
 
 #ifndef NFS_MAXHOSTNAME
@@ -65,79 +62,90 @@
 #define NFS_MAXPATHNAME		(1024)
 #endif
 
+#ifndef NFS_DEF_FG_TIMEOUT_MINUTES
+#define NFS_DEF_FG_TIMEOUT_MINUTES	(2u)
+#endif
+
+#ifndef NFS_DEF_BG_TIMEOUT_MINUTES
+#define NFS_DEF_BG_TIMEOUT_MINUTES	(10000u)
+#endif
+
 extern int nfs_mount_data_version;
 extern char *progname;
 extern int verbose;
+extern int sloppy;
 
-static int parse_devname(const char *spec, char **hostname)
+struct nfsmount_info {
+	const char		*spec,		/* server:/path */
+				*node,		/* mounted-on dir */
+				*type;		/* "nfs" or "nfs4" */
+	char			*hostname;	/* server's hostname */
+
+	struct mount_options	*options;	/* parsed mount options */
+	char			**extra_opts;	/* string for /etc/mtab */
+
+	int			flags,		/* MS_ flags */
+				fake,		/* actually do the mount? */
+				child;		/* forked bg child? */
+
+	sa_family_t		family;		/* supported address family */
+};
+
+/*
+ * Obtain a retry timeout value based on the value of the "retry=" option.
+ *
+ * Returns a time_t timeout timestamp, in seconds.
+ */
+static time_t nfs_parse_retry_option(struct mount_options *options,
+				     unsigned int timeout_minutes)
 {
-	int ret = 0;
-	char *dev, *pathname, *s;
+	char *retry_option, *endptr;
 
-	dev = xstrdup(spec);
+	retry_option = po_get(options, "retry");
+	if (retry_option) {
+		long tmp;
 
-	if (!(pathname = strchr(dev, ':'))) {
-		nfs_error(_("%s: remote share not in 'host:dir' format"),
-				progname);
-		goto out;
-	}
-	*pathname = '\0';
-	pathname++;
-
-	/*
-	 * We don't need a copy of the pathname, but let's
-	 * sanity check it anyway.
-	 */
-	if (strlen(pathname) > NFS_MAXPATHNAME) {
-		nfs_error(_("%s: export pathname is too long"),
-				progname);
-		goto out;
+		errno = 0;
+		tmp = strtol(retry_option, &endptr, 10);
+		if (errno == 0 && endptr != retry_option && tmp >= 0)
+			timeout_minutes = tmp;
+		else if (verbose)
+			nfs_error(_("%s: invalid retry timeout was specified; "
+					"using default timeout"), progname);
 	}
 
-	/*
-	 * Ignore all but first hostname in replicated mounts
-	 * until they can be fully supported. (mack@sgi.com)
-	 */
-	if ((s = strchr(dev, ','))) {
-		*s = '\0';
-		nfs_error(_("%s: warning: multiple hostnames not supported"),
-				progname);
-		nfs_error(_("%s: ignoring hostnames that follow the first one"),
-				progname);
-	}
-	*hostname = xstrdup(dev);
-	if (strlen(*hostname) > NFS_MAXHOSTNAME) {
-		nfs_error(_("%s: server hostname is too long"),
-				progname);
-		free(*hostname);
-		goto out;
-	}
-
-	ret = 1;
-
-out:
-	free(dev);
-	return ret;
+	return time(NULL) + (time_t)(timeout_minutes * 60);
 }
 
-static int fill_ipv4_sockaddr(const char *hostname, struct sockaddr_in *addr)
+/*
+ * Convert the passed-in sockaddr-style address to presentation
+ * format, then append an option of the form "keyword=address".
+ *
+ * Returns 1 if the option was appended successfully; otherwise zero.
+ */
+static int nfs_append_generic_address_option(const struct sockaddr *sap,
+					     const socklen_t salen,
+					     const char *keyword,
+					     struct mount_options *options)
 {
-	struct hostent *hp;
-	addr->sin_family = AF_INET;
+	char address[NI_MAXHOST];
+	char new_option[512];
+
+	if (!nfs_present_sockaddr(sap, salen, address, sizeof(address)))
+		goto out_err;
+
+	if (snprintf(new_option, sizeof(new_option), "%s=%s",
+					keyword, address) >= sizeof(new_option))
+		goto out_err;
+
+	if (po_append(options, new_option) != PO_SUCCEEDED)
+		goto out_err;
 
-	if (inet_aton(hostname, &addr->sin_addr))
-		return 1;
-	if ((hp = gethostbyname(hostname)) == NULL) {
-		nfs_error(_("%s: can't get address for %s\n"),
-				progname, hostname);
-		return 0;
-	}
-	if (hp->h_length > sizeof(struct in_addr)) {
-		nfs_error(_("%s: got bad hp->h_length"), progname);
-		hp->h_length = sizeof(struct in_addr);
-	}
-	memcpy(&addr->sin_addr, hp->h_addr, hp->h_length);
 	return 1;
+
+out_err:
+	nfs_error(_("%s: failed to construct %s option"), progname, keyword);
+	return 0;
 }
 
 /*
@@ -152,19 +160,12 @@ static int fill_ipv4_sockaddr(const char *hostname, struct sockaddr_in *addr)
  * Returns 1 if 'addr=' option appended successfully;
  * otherwise zero.
  */
-static int append_addr_option(struct sockaddr_in *saddr,
-			   struct mount_options *options)
+static int nfs_append_addr_option(const struct sockaddr *sap,
+				  socklen_t salen,
+				  struct mount_options *options)
 {
-	char new_option[24];
-
 	po_remove_all(options, "addr");
-
-	snprintf(new_option, sizeof(new_option) - 1,
-			"addr=%s", inet_ntoa(saddr->sin_addr));
-
-	if (po_append(options, new_option) == PO_SUCCEEDED)
-		return 1;
-	return 0;
+	return nfs_append_generic_address_option(sap, salen, "addr", options);
 }
 
 /*
@@ -174,73 +175,112 @@ static int append_addr_option(struct sockaddr_in *saddr,
  * Returns 1 if 'clientaddr=' option created successfully or if
  * 'clientaddr=' option is already present; otherwise zero.
  */
-static int append_clientaddr_option(struct sockaddr_in *saddr,
-				    struct mount_options *options)
+static int nfs_append_clientaddr_option(const struct sockaddr *sap,
+					socklen_t salen,
+					struct mount_options *options)
 {
-	struct sockaddr_in my_addr;
-	char new_option[32];
+	struct sockaddr_storage dummy;
+	struct sockaddr *my_addr = (struct sockaddr *)&dummy;
+	socklen_t my_len = sizeof(dummy);
 
-	if (po_contains(options, "clientaddr") == PO_SUCCEEDED)
+	if (po_contains(options, "clientaddr") == PO_FOUND)
 		return 1;
 
-	if (!get_client_address(saddr, &my_addr))
-		return 0;
+	nfs_callback_address(sap, salen, my_addr, &my_len);
 
-	snprintf(new_option, sizeof(new_option) - 1,
-			"clientaddr=%s", inet_ntoa(my_addr.sin_addr));
-
-	if (po_append(options, new_option) == PO_SUCCEEDED)
-		return 1;
-	return 0;
+	return nfs_append_generic_address_option(my_addr, my_len,
+							"clientaddr", options);
 }
 
 /*
  * Resolve the 'mounthost=' hostname and append a new option using
- * the resulting IPv4 address.
+ * the resulting address.
  */
-static int fix_mounthost_option(struct mount_options *options)
+static int nfs_fix_mounthost_option(const sa_family_t family,
+				    struct mount_options *options)
 {
-	struct sockaddr_in maddr;
-	char *mounthost, new_option[32];
+	struct sockaddr_storage dummy;
+	struct sockaddr *sap = (struct sockaddr *)&dummy;
+	socklen_t salen = sizeof(dummy);
+	char *mounthost;
 
 	mounthost = po_get(options, "mounthost");
 	if (!mounthost)
 		return 1;
 
-	if (!fill_ipv4_sockaddr(mounthost, &maddr))
+	if (!nfs_name_to_address(mounthost, family, sap, &salen)) {
+		nfs_error(_("%s: unable to determine mount server's address"),
+				progname);
 		return 0;
+	}
 
-	snprintf(new_option, sizeof(new_option) - 1,
-			"mountaddr=%s", inet_ntoa(maddr.sin_addr));
-
-	if (po_append(options, new_option) == PO_SUCCEEDED)
-		return 1;
-	return 0;
+	return nfs_append_generic_address_option(sap, salen,
+							"mountaddr", options);
 }
 
 /*
- * Set up mandatory mount options.
- *
- * Returns 1 if successful; otherwise zero.
+ * Returns zero if the "lock" option is in effect, but statd
+ * can't be started.  Otherwise, returns 1.
  */
-static int set_mandatory_options(const char *type,
-				 struct sockaddr_in *saddr,
-				 struct mount_options *options)
+static int nfs_verify_lock_option(struct mount_options *options)
 {
-	if (!append_addr_option(saddr, options))
-		return 0;
+	if (po_rightmost(options, "nolock", "lock") == PO_KEY1_RIGHTMOST)
+		return 1;
 
-	if (strncmp(type, "nfs4", 4) == 0) {
-		if (!append_clientaddr_option(saddr, options))
-			return 0;
-	} else {
-		if (!fix_mounthost_option(options))
-			return 0;
+	if (!start_statd()) {
+		nfs_error(_("%s: rpc.statd is not running but is "
+			    "required for remote locking."), progname);
+		nfs_error(_("%s: Either use '-o nolock' to keep "
+			    "locks local, or start statd."), progname);
+		return 0;
 	}
 
 	return 1;
 }
 
+static int nfs_append_sloppy_option(struct mount_options *options)
+{
+	if (!sloppy || linux_version_code() < MAKE_VERSION(2, 6, 27))
+		return 1;
+
+	if (po_append(options, "sloppy") == PO_FAILED)
+		return 0;
+	return 1;
+}
+
+/*
+ * Set up mandatory NFS mount options.
+ *
+ * Returns 1 if successful; otherwise zero.
+ */
+static int nfs_validate_options(struct nfsmount_info *mi)
+{
+	struct sockaddr_storage dummy;
+	struct sockaddr *sap = (struct sockaddr *)&dummy;
+	socklen_t salen = sizeof(dummy);
+
+	if (!nfs_parse_devname(mi->spec, &mi->hostname, NULL))
+		return 0;
+
+	if (!nfs_name_to_address(mi->hostname, mi->family, sap, &salen))
+		return 0;
+
+	if (strncmp(mi->type, "nfs4", 4) == 0) {
+		if (!nfs_append_clientaddr_option(sap, salen, mi->options))
+			return 0;
+	} else {
+		if (!nfs_fix_mounthost_option(mi->family, mi->options))
+			return 0;
+		if (!mi->fake && !nfs_verify_lock_option(mi->options))
+			return 0;
+	}
+
+	if (!nfs_append_sloppy_option(mi->options))
+		return 0;
+
+	return nfs_append_addr_option(sap, salen, mi->options);
+}
+
 /*
  * Distinguish between permanent and temporary errors.
  *
@@ -249,10 +289,9 @@ static int set_mandatory_options(const char *type,
  * passed-in error is permanent, thus the mount system call
  * should not be retried.
  */
-static int is_permanent_error(int error)
+static int nfs_is_permanent_error(int error)
 {
 	switch (error) {
-	case EACCES:
 	case ESTALE:
 	case ETIMEDOUT:
 	case ECONNREFUSED:
@@ -269,13 +308,13 @@ static int is_permanent_error(int error)
  *
  * To handle version and transport protocol fallback properly, we
  * need to parse some of the mount options in order to set up a
- * portmap probe.  Mount options that rewrite_mount_options()
+ * portmap probe.  Mount options that nfs_rewrite_mount_options()
  * doesn't recognize are left alone.
  *
  * Returns a new group of mount options if successful; otherwise
  * NULL is returned if some failure occurred.
  */
-static struct mount_options *rewrite_mount_options(char *str)
+static struct mount_options *nfs_rewrite_mount_options(char *str)
 {
 	struct mount_options *options;
 	char *option, new_option[64];
@@ -284,9 +323,12 @@ static struct mount_options *rewrite_mount_options(char *str)
 	int p;
 
 	options = po_split(str);
-	if (!options)
+	if (!options) {
+		errno = EFAULT;
 		return NULL;
+	}
 
+	errno = EINVAL;
 	option = po_get(options, "addr");
 	if (option) {
 		nfs_server.saddr.sin_family = AF_INET;
@@ -311,6 +353,17 @@ static struct mount_options *rewrite_mount_options(char *str)
 	option = po_get(options, "mountvers");
 	if (option)
 		mnt_server.pmap.pm_vers = atoi(option);
+	option = po_get(options, "mountproto");
+	if (option) {
+		if (strcmp(option, "tcp") == 0) {
+			mnt_server.pmap.pm_prot = IPPROTO_TCP;
+			po_remove_all(options, "mountproto");
+		}
+		if (strcmp(option, "udp") == 0) {
+			mnt_server.pmap.pm_prot = IPPROTO_UDP;
+			po_remove_all(options, "mountproto");
+		}
+	}
 
 	option = po_get(options, "port");
 	if (option) {
@@ -353,7 +406,7 @@ static struct mount_options *rewrite_mount_options(char *str)
 	po_remove_all(options, "udp");
 
 	if (!probe_bothports(&mnt_server, &nfs_server)) {
-		rpc_mount_errors("rpcbind", 0, 0);
+		errno = ESPIPE;
 		goto err;
 	}
 
@@ -379,6 +432,21 @@ static struct mount_options *rewrite_mount_options(char *str)
 
 	}
 
+	if (mnt_server.pmap.pm_prot == IPPROTO_TCP)
+		snprintf(new_option, sizeof(new_option) - 1,
+			 "mountproto=tcp");
+	else
+		snprintf(new_option, sizeof(new_option) - 1,
+			 "mountproto=udp");
+	if (po_append(options, new_option) == PO_FAILED)
+		goto err;
+
+	snprintf(new_option, sizeof(new_option) - 1,
+		 "mountport=%lu", mnt_server.pmap.pm_port);
+	if (po_append(options, new_option) == PO_FAILED)
+		goto err;
+
+	errno = 0;
 	return options;
 
 err:
@@ -386,6 +454,27 @@ err:
 	return NULL;
 }
 
+/*
+ * Do the mount(2) system call.
+ *
+ * Returns 1 if successful, otherwise zero.
+ * "errno" is set to reflect the individual error.
+ */
+static int nfs_sys_mount(const struct nfsmount_info *mi, const char *type,
+			 const char *options)
+{
+	int result;
+
+	result = mount(mi->spec, mi->node, type,
+				mi->flags & ~(MS_USER|MS_USERS), options);
+	if (verbose && result) {
+		int save = errno;
+		nfs_error(_("%s: mount(2): %s"), progname, strerror(save));
+		errno = save;
+	}
+	return !result;
+}
+
 /*
  * Retry an NFS mount that failed because the requested service isn't
  * available on the server.
@@ -397,18 +486,15 @@ err:
  * 'extra_opts' are updated to reflect the mount options that worked.
  * If the retry fails, 'options' and 'extra_opts' are left unchanged.
  */
-static int retry_nfsmount(const char *spec, const char *node,
-			int flags, struct mount_options *options,
-			int fake, char **extra_opts)
+static int nfs_retry_nfs23mount(struct nfsmount_info *mi)
 {
 	struct mount_options *retry_options;
 	char *retry_str = NULL;
+	char **extra_opts = mi->extra_opts;
 
-	retry_options = rewrite_mount_options(*extra_opts);
-	if (!retry_options) {
-		errno = EIO;
+	retry_options = nfs_rewrite_mount_options(*extra_opts);
+	if (!retry_options)
 		return 0;
-	}
 
 	if (po_join(retry_options, &retry_str) == PO_FAILED) {
 		po_destroy(retry_options);
@@ -420,17 +506,16 @@ static int retry_nfsmount(const char *spec, const char *node,
 		printf(_("%s: text-based options (retry): '%s'\n"),
 			progname, retry_str);
 
-	if (!mount(spec, node, "nfs",
-				flags & ~(MS_USER|MS_USERS), retry_str)) {
-		free(*extra_opts);
-		*extra_opts = retry_str;
-		po_replace(options, retry_options);
-		return 1;
+	if (!nfs_sys_mount(mi, "nfs", retry_str)) {
+		po_destroy(retry_options);
+		free(retry_str);
+		return 0;
 	}
 
-	po_destroy(retry_options);
-	free(retry_str);
-	return 0;
+	free(*extra_opts);
+	*extra_opts = retry_str;
+	po_replace(mi->options, retry_options);
+	return 1;
 }
 
 /*
@@ -446,11 +531,11 @@ static int retry_nfsmount(const char *spec, const char *node,
  * 'extra_opts' are updated to reflect the mount options that worked.
  * If the retry fails, 'options' and 'extra_opts' are left unchanged.
  */
-static int try_nfs23mount(const char *spec, const char *node,
-			  int flags, struct mount_options *options,
-			  int fake, char **extra_opts)
+static int nfs_try_nfs23mount(struct nfsmount_info *mi)
 {
-	if (po_join(options, extra_opts) == PO_FAILED) {
+	char **extra_opts = mi->extra_opts;
+
+	if (po_join(mi->options, extra_opts) == PO_FAILED) {
 		errno = EIO;
 		return 0;
 	}
@@ -459,11 +544,10 @@ static int try_nfs23mount(const char *spec, const char *node,
 		printf(_("%s: text-based options: '%s'\n"),
 			progname, *extra_opts);
 
-	if (fake)
+	if (mi->fake)
 		return 1;
 
-	if (!mount(spec, node, "nfs",
-				flags & ~(MS_USER|MS_USERS), *extra_opts))
+	if (nfs_sys_mount(mi, "nfs", *extra_opts))
 		return 1;
 
 	/*
@@ -473,7 +557,7 @@ static int try_nfs23mount(const char *spec, const char *node,
 	if (errno != EOPNOTSUPP && errno != EPROTONOSUPPORT)
 		return 0;
 
-	return retry_nfsmount(spec, node, flags, options, fake, extra_opts);
+	return nfs_retry_nfs23mount(mi);
 }
 
 /*
@@ -482,11 +566,11 @@ static int try_nfs23mount(const char *spec, const char *node,
  * Returns 1 if successful.  Otherwise, returns zero.
  * "errno" is set to reflect the individual error.
  */
-static int try_nfs4mount(const char *spec, const char *node,
-			 int flags, struct mount_options *options,
-			 int fake, char **extra_opts)
+static int nfs_try_nfs4mount(struct nfsmount_info *mi)
 {
-	if (po_join(options, extra_opts) == PO_FAILED) {
+	char **extra_opts = mi->extra_opts;
+
+	if (po_join(mi->options, extra_opts) == PO_FAILED) {
 		errno = EIO;
 		return 0;
 	}
@@ -495,31 +579,24 @@ static int try_nfs4mount(const char *spec, const char *node,
 		printf(_("%s: text-based options: '%s'\n"),
 			progname, *extra_opts);
 
-	if (fake)
+	if (mi->fake)
 		return 1;
 
-	if (!mount(spec, node, "nfs4",
-				flags & ~(MS_USER|MS_USERS), *extra_opts))
-		return 1;
-	return 0;
+	return nfs_sys_mount(mi, "nfs4", *extra_opts);
 }
 
 /*
- * Try the mount(2) system call.
+ * Perform either an NFSv2/3 mount, or an NFSv4 mount system call.
  *
  * Returns 1 if successful.  Otherwise, returns zero.
  * "errno" is set to reflect the individual error.
  */
-static int try_mount(const char *spec, const char *node, const char *type,
-		     int flags, struct mount_options *options, int fake,
-		     char **extra_opts)
+static int nfs_try_mount(struct nfsmount_info *mi)
 {
-	if (strncmp(type, "nfs4", 4) == 0)
-		return try_nfs4mount(spec, node, flags,
-					options, fake, extra_opts);
+	if (strncmp(mi->type, "nfs4", 4) == 0)
+		return nfs_try_nfs4mount(mi);
 	else
-		return try_nfs23mount(spec, node, flags,
-					options, fake, extra_opts);
+		return nfs_try_nfs23mount(mi);
 }
 
 /*
@@ -529,30 +606,22 @@ static int try_mount(const char *spec, const char *node, const char *type,
  *
  * Returns a valid mount command exit code.
  */
-static int nfsmount_fg(const char *spec, const char *node,
-		       const char *type, int flags,
-		       struct mount_options *options, int fake,
-		       char **extra_opts)
+static int nfsmount_fg(struct nfsmount_info *mi)
 {
 	unsigned int secs = 1;
-	time_t timeout = time(NULL);
-	char *retry;
-
-	timeout += 60 * 2;		/* default: 2 minutes */
-	retry = po_get(options, "retry");
-	if (retry)
-		timeout += 60 * atoi(retry);
+	time_t timeout;
 
+	timeout = nfs_parse_retry_option(mi->options,
+					 NFS_DEF_FG_TIMEOUT_MINUTES);
 	if (verbose)
 		printf(_("%s: timeout set for %s"),
 			progname, ctime(&timeout));
 
 	for (;;) {
-		if (try_mount(spec, node, type, flags,
-					options, fake, extra_opts))
+		if (nfs_try_mount(mi))
 			return EX_SUCCESS;
 
-		if (is_permanent_error(errno))
+		if (nfs_is_permanent_error(errno))
 			break;
 
 		if (time(NULL) > timeout) {
@@ -569,7 +638,7 @@ static int nfsmount_fg(const char *spec, const char *node,
 		}
 	};
 
-	mount_error(spec, node, errno);
+	mount_error(mi->spec, mi->node, errno);
 	return EX_FAIL;
 }
 
@@ -580,21 +649,17 @@ static int nfsmount_fg(const char *spec, const char *node,
  *
  * EX_BG should cause the caller to fork and invoke nfsmount_child.
  */
-static int nfsmount_parent(const char *spec, const char *node,
-			   const char *type, char *hostname, int flags,
-			   struct mount_options *options,
-			   int fake, char **extra_opts)
+static int nfsmount_parent(struct nfsmount_info *mi)
 {
-	if (try_mount(spec, node, type, flags, options,
-					fake, extra_opts))
+	if (nfs_try_mount(mi))
 		return EX_SUCCESS;
 
-	if (is_permanent_error(errno)) {
-		mount_error(spec, node, errno);
+	if (nfs_is_permanent_error(errno)) {
+		mount_error(mi->spec, mi->node, errno);
 		return EX_FAIL;
 	}
 
-	sys_mount_errors(hostname, errno, 1, 1);
+	sys_mount_errors(mi->hostname, errno, 1, 1);
 	return EX_BG;
 }
 
@@ -606,19 +671,13 @@ static int nfsmount_parent(const char *spec, const char *node,
  * error return, though, so we use sys_mount_errors to log the
  * failure.
  */
-static int nfsmount_child(const char *spec, const char *node,
-			  const char *type, char *hostname, int flags,
-			  struct mount_options *options,
-			  int fake, char **extra_opts)
+static int nfsmount_child(struct nfsmount_info *mi)
 {
 	unsigned int secs = 1;
-	time_t timeout = time(NULL);
-	char *retry;
+	time_t timeout;
 
-	timeout += 60 * 10000;		/* default: 10,000 minutes */
-	retry = po_get(options, "retry");
-	if (retry)
-		timeout += 60 * atoi(retry);
+	timeout = nfs_parse_retry_option(mi->options,
+					 NFS_DEF_BG_TIMEOUT_MINUTES);
 
 	for (;;) {
 		if (sleep(secs))
@@ -627,20 +686,19 @@ static int nfsmount_child(const char *spec, const char *node,
 		if (secs > 120)
 			secs = 120;
 
-		if (try_mount(spec, node, type, flags, options,
-							fake, extra_opts))
+		if (nfs_try_mount(mi))
 			return EX_SUCCESS;
 
-		if (is_permanent_error(errno))
+		if (nfs_is_permanent_error(errno))
 			break;
 
 		if (time(NULL) > timeout)
 			break;
 
-		sys_mount_errors(hostname, errno, 1, 1);
+		sys_mount_errors(mi->hostname, errno, 1, 1);
 	};
 
-	sys_mount_errors(hostname, errno, 1, 0);
+	sys_mount_errors(mi->hostname, errno, 1, 0);
 	return EX_FAIL;
 }
 
@@ -649,17 +707,28 @@ static int nfsmount_child(const char *spec, const char *node,
  *
  * Returns a valid mount command exit code.
  */
-static int nfsmount_bg(const char *spec, const char *node,
-			  const char *type, char *hostname, int flags,
-			  struct mount_options *options,
-			  int fake, int child, char **extra_opts)
+static int nfsmount_bg(struct nfsmount_info *mi)
 {
-	if (!child)
-		return nfsmount_parent(spec, node, type, hostname, flags,
-					options, fake, extra_opts);
+	if (!mi->child)
+		return nfsmount_parent(mi);
 	else
-		return nfsmount_child(spec, node, type, hostname, flags,
-					options, fake, extra_opts);
+		return nfsmount_child(mi);
+}
+
+/*
+ * Process mount options and try a mount system call.
+ *
+ * Returns a valid mount command exit code.
+ */
+static int nfsmount_start(struct nfsmount_info *mi)
+{
+	if (!nfs_validate_options(mi))
+		return EX_FAIL;
+
+	if (po_rightmost(mi->options, "bg", "fg") == PO_KEY1_RIGHTMOST)
+		return nfsmount_bg(mi);
+	else
+		return nfsmount_fg(mi);
 }
 
 /**
@@ -676,35 +745,29 @@ static int nfsmount_bg(const char *spec, const char *node,
 int nfsmount_string(const char *spec, const char *node, const char *type,
 		    int flags, char **extra_opts, int fake, int child)
 {
-	struct mount_options *options = NULL;
-	struct sockaddr_in saddr;
-	char *hostname;
+	struct nfsmount_info mi = {
+		.spec		= spec,
+		.node		= node,
+		.type		= type,
+		.extra_opts	= extra_opts,
+		.flags		= flags,
+		.fake		= fake,
+		.child		= child,
+#ifdef IPV6_SUPPORTED
+		.family		= AF_UNSPEC,	/* either IPv4 or v6 */
+#else
+		.family		= AF_INET,	/* only IPv4 */
+#endif
+	};
 	int retval = EX_FAIL;
 
-	if (!parse_devname(spec, &hostname))
-		return retval;
-	if (!fill_ipv4_sockaddr(hostname, &saddr))
-		goto fail;
-
-	options = po_split(*extra_opts);
-	if (!options) {
+	mi.options = po_split(*extra_opts);
+	if (mi.options) {
+		retval = nfsmount_start(&mi);
+		po_destroy(mi.options);
+	} else
 		nfs_error(_("%s: internal option parsing error"), progname);
-		goto fail;
-	}
 
-	if (!set_mandatory_options(type, &saddr, options))
-		goto out;
-
-	if (po_rightmost(options, "bg", "fg") == PO_KEY1_RIGHTMOST)
-		retval = nfsmount_bg(spec, node, type, hostname, flags,
-					options, fake, child, extra_opts);
-	else
-		retval = nfsmount_fg(spec, node, type, flags, options,
-							fake, extra_opts);
-
-out:
-	po_destroy(options);
-fail:
-	free(hostname);
+	free(mi.hostname);
 	return retval;
 }
