@@ -12,7 +12,9 @@
  *	260295	 okir	started with simply syslog logging.
  */
 
-#include "config.h"
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
 
 #include <syslog.h>
 #include <unistd.h>
@@ -27,9 +29,11 @@
 
 static pid_t	mypid;
 								/* Turns on logging to console/stderr. */
+#if 0
 static int	opt_debug = 0;	/* Will be command-line option, eventually */
+#endif
 
-void log_init()
+void log_init(void)
 {
 	if (!(run_mode & MODE_LOG_STDERR)) 
 		openlog(name_p, LOG_PID | LOG_NDELAY, LOG_DAEMON);
